"""Support testing with Pytest."""

import logging
import os
from contextlib import AsyncExitStack

import pytest
from asgi_tools.tests import ASGITestClient, manage_lifespan


def pytest_addoption(parser):
    """Append pytest options for testing Muffin apps."""
    parser.addini("muffin_app", "Set path to muffin application")
    parser.addoption(
        "--muffin-app",
        dest="muffin_app",
        help="Set to muffin application",
    )

    parser.addini("muffin_config", "Set module path to muffin configuration")
    parser.addoption(
        "--muffin-config",
        dest="muffin_config",
        help="Set module path to muffin configuration",
    )


def pytest_load_initial_conftests(early_config, parser, args):
    """Prepare to loading Muffin application."""
    from muffin.constants import CONFIG_ENV_VARIABLE

    options = parser.parse_known_args(args)

    # Initialize configuration
    config = options.muffin_config or early_config.getini("muffin_config")
    if config:
        os.environ[CONFIG_ENV_VARIABLE] = config

    # Initialize application
    app_ = options.muffin_app or early_config.getini("muffin_app")
    early_config.app = app_


@pytest.fixture(scope="session")
async def app(pytestconfig, request, aiolib):  # noqa: ARG001
    """Load an application, run lifespan events, prepare plugins."""
    if not pytestconfig.app:
        logging.warning(
            (
                "Improperly configured. Please set ``muffin_app`` in your pytest config. "
                "Or use ``--muffin-app`` command option."
            ),
        )
        return

    from muffin.utils import import_app

    app_ = import_app(pytestconfig.app)
    msg = f"Setup application '{app_.cfg.name}'"
    if app_.cfg.config:
        msg += f"with config '{app_.cfg.config}'"
    app_.logger.info(msg)

    # Setup plugins
    async with AsyncExitStack() as stack:
        for plugin in app_.plugins.values():
            conftest = getattr(plugin, "conftest", None)
            if conftest:
                app_.logger.info("Setup plugin '%s'", plugin.name)
                await stack.enter_async_context(conftest())

        # Manage lifespan
        async with manage_lifespan(app_):
            yield app_


@pytest.fixture()
def client(app):
    """Generate a test client for the app."""
    return ASGITestClient(app)



import pytest


@pytest.fixture()
def name(app):
    return app.cfg.name


def test_app_imported(app):
    assert app.cfg.name == "muffin"


def test_app_available_in_fixture(name):
    assert name == "muffin"


def test_app_lifespan(app):
    assert app.state == "started"



"""The Muffin Utils."""

from __future__ import annotations

import asyncio
import importlib
import os
import pkgutil
import sys
import threading
from collections import OrderedDict
from contextlib import suppress
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, TypeVar

from asgi_tools.utils import is_awaitable, to_awaitable
from sniffio import current_async_library

from muffin.errors import InvalidAppError

if TYPE_CHECKING:
    from types import ModuleType

    from asgi_tools.types import TASGIApp

__all__ = (
    "aio_lib",
    "aio_run",
    "import_submodules",
    "current_async_library",
    "is_awaitable",
    "to_awaitable",
)

AIOLIB = threading.local()
AIOLIB.current = None
AIOLIBS: Dict[str, ModuleType] = OrderedDict()

with suppress(ImportError):
    import curio

    AIOLIBS["curio"] = curio

with suppress(ImportError):
    import trio

    AIOLIBS["trio"] = trio

AIOLIBS["asyncio"] = asyncio

TV = TypeVar("TV")


def aio_lib() -> str:
    """Return first available async library."""
    aiolib = os.environ.get("MUFFIN_AIOLIB", "asyncio")
    if aiolib:
        return aiolib

    for name, module in AIOLIBS.items():
        if module is not None:
            return name

    return "asyncio"


def aio_run(corofn: Callable[..., Coroutine[None, None, TV]], *args, **kwargs) -> TV:
    """Run the given coroutine with current async library."""
    AIOLIB.current = aiolib = AIOLIB.current or aio_lib()
    if aiolib == "asyncio":
        return asyncio.run(corofn(*args, **kwargs))

    return AIOLIBS[aiolib].run(lambda: corofn(*args, **kwargs))


def import_submodules(package_name: str, *submodules: str) -> Dict[str, ModuleType]:
    """Import all submodules by package name."""
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + "." + name)
        for _, name, _ in pkgutil.walk_packages(package.__path__)
        if not submodules or name in submodules
    }


def import_app(app_uri: str, *, reload: bool = False) -> TASGIApp:
    """Import application by the given string (python.path.to.module:app_name)."""
    mod_name, _, app_name = app_uri.partition(":")
    mod = importlib.import_module(mod_name)
    if reload:
        importlib.reload(mod)

    try:
        return getattr(mod, app_name or "app", None) or mod.application
    except AttributeError as exc:
        raise InvalidAppError(app_uri) from exc



"""Muffin is a web framework to build ASGI applications."""
from __future__ import annotations

from asgi_tools import (
    ASGIConnectionClosedError,
    ASGIError,
    ASGIInvalidMethodError,
    ASGINotFoundError,
    Request,
    Response,
    ResponseError,
    ResponseFile,
    ResponseHTML,
    ResponseJSON,
    ResponseRedirect,
    ResponseSSE,
    ResponseStream,
    ResponseText,
    ResponseWebSocket,
)
from asgi_tools.tests import ASGITestClient as TestClient

from .app import Application
from .errors import MuffinError
from .handler import Handler

__all__ = (
    "ASGIConnectionClosedError",
    "ASGIError",
    "ASGIInvalidMethodError",
    "ASGINotFoundError",
    "Application",
    "Handler",
    "MuffinError",
    "Request",
    "Response",
    "ResponseError",
    "ResponseFile",
    "ResponseHTML",
    "ResponseJSON",
    "ResponseRedirect",
    "ResponseSSE",
    "ResponseStream",
    "ResponseText",
    "ResponseWebSocket",
    "TestClient",
)



def test_imports():
    import muffin

    assert hasattr(muffin, "Request")
    assert hasattr(muffin, "Response")
    assert hasattr(muffin, "ResponseError")
    assert hasattr(muffin, "ResponseFile")
    assert hasattr(muffin, "ResponseHTML")
    assert hasattr(muffin, "ResponseJSON")
    assert hasattr(muffin, "ResponseRedirect")
    assert hasattr(muffin, "ResponseStream")
    assert hasattr(muffin, "ResponseText")
    assert hasattr(muffin, "ASGINotFoundError")
    assert hasattr(muffin, "ASGIInvalidMethodError")
    assert hasattr(muffin, "TestClient")



"""Base Tests."""
from __future__ import annotations

from pathlib import Path
from unittest import mock


def test_app_config():
    import os

    import muffin

    os.environ["TEST_DEBUG"] = "true"

    app = muffin.Application(
        "tests.config_fixture",
        config="unknown",
        name="test",
        LOG_CONFIG={"version": 1},
    )
    assert app.cfg
    assert app.cfg.CONFIG == "tests.config_fixture"
    assert app.cfg.CONFIG_VARIABLE == 42
    assert app.cfg.DEBUG is True
    assert app.cfg.MANAGE_SHELL
    assert app.cfg.STATIC_URL_PREFIX == "/static"


def test_configure_logging():
    import muffin

    dummy = {"dummy": "dict", "version": 1}
    with mock.patch("muffin.app.dictConfig") as mocked:
        app = muffin.Application("muffin", LOG_CONFIG=dummy)
        assert app.logger
        assert app.logger.handlers
    mocked.assert_called_once_with(dummy)


async def test_static_folders():
    import muffin

    app = muffin.Application(
        static_folders=["tests", Path(__file__).parent.parent],
        static_url_prefix="/assets",
    )
    assert app.cfg.STATIC_FOLDERS
    assert app.cfg.STATIC_URL_PREFIX == "/assets"

    @app.route("/")
    async def index(request):
        return "OK"

    client = muffin.TestClient(app)
    res = await client.get("/")
    assert res.status_code == 200

    res = await client.get("/assets/test_application.py")
    assert res.status_code == 200
    text = await res.text()
    assert text.startswith('"""Base Tests."""')

    res = await client.get("/assets/pyproject.toml")
    assert res.status_code == 200


# ruff: noqa: ARG001
